# -*- coding: utf-8 -*-
# (c) 2024, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
import argparse
import re
from io import StringIO
from pathlib import Path

from .base import AndeboxAction


def info_type(types, v):
    try:
        r = [t for t in types if t.startswith(v.lower())]
        return r[0][0].upper()
    except IndexError as e:
        raise argparse.ArgumentTypeError("invalid value: {v}") from e


class ReformatYAMLAction(AndeboxAction):
    name = "reformat-yaml"
    help = "reformat YAML content in plugins"
    args = [
        dict(
            names=("files",),
            specs=dict(
                help="Python files where to search for YAML content",
                type=Path,
                nargs="+",
            ),
        ),
    ]

    @staticmethod
    def format_yaml(yaml_content: str) -> str:
        from ruamel.yaml import YAML

        yaml = YAML()
        yaml.preserve_quotes = True
        yaml.indent(2)
        yaml.explicit_start = True
        yaml.width = 160
        yaml.preserve_quotes = True
        yaml.top_level_colon_align = False
        yaml.compact_seq_seq = False

        data = yaml.load(yaml_content)
        output = StringIO()
        yaml.dump(data, output)
        return output.getvalue(), data

    def reformat_yaml_in_python_file(self, file_path):
        QUOTE_RE_FRAG = r'(?:"|\'){3}'
        VAR_RE_FRAG = r"[A-Z][A-Z_0-9]*"
        ONELINE_RE = re.compile(
            rf"^\s*{VAR_RE_FRAG}\s*=\s*r?{QUOTE_RE_FRAG}.*{QUOTE_RE_FRAG}$"
        )
        FIRST_LINE_RE = re.compile(rf"^(\s*{VAR_RE_FRAG})\s*=\s*r?{QUOTE_RE_FRAG}(.*)$")

        print(f"Opening file {file_path}")
        with open(file_path, "r") as file:
            lines = file.readlines()

        updated_lines = []

        variable = ""
        first_line = ""
        quoted_content = []

        for line in lines:
            line = line.rstrip()
            if variable:
                if not re.match(rf"^\s*{QUOTE_RE_FRAG}", line):
                    quoted_content.append(line)
                    continue

                updated_lines.append(first_line)

                yaml_content, data = self.format_yaml("\n".join(quoted_content))
                if data:
                    yaml_content = yaml_content.splitlines()
                    while yaml_content[0] == "":
                        yaml_content = yaml_content[1:]
                    updated_lines.extend(yaml_content)
                else:
                    # If no data (e.g. an empty or comment-only block), then keep original content
                    updated_lines.extend(quoted_content)

                updated_lines.append('"""')

                variable = ""
                quoted_content = []

            else:
                if ONELINE_RE.search(line):
                    updated_lines.append(re.sub(QUOTE_RE_FRAG, '"""', line))
                elif match := FIRST_LINE_RE.search(line):
                    variable, yaml_first_line = match.groups()
                    first_line = f'{variable} = """{yaml_first_line}'
                else:
                    updated_lines.append(line)

        # Write the updated lines back to the file
        with open(file_path, "w") as file:
            file.writelines([f"{x}\n" for x in updated_lines])

    def run(self, context):
        for file_path in context.args.files:
            self.reformat_yaml_in_python_file(file_path)



import os
from contextlib import contextmanager
from pathlib import Path


@contextmanager
def set_dir(path):
    previous = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(previous)



# -*- coding: utf-8 -*-
# (c) 2023, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
import os
import shutil
import subprocess
import sys
import tempfile
from abc import ABC
from abc import abstractmethod
from contextlib import contextmanager
from pathlib import Path

import yaml

from .exceptions import AndeboxException

toplevel_exclusion = (".git", ".tox", ".venv", ".virtualvenv", "venv", "virtualenv")


class AndeboxUnknownContext(AndeboxException):
    pass


class AbstractContext(ABC):

    def __init__(self, base_dir, args) -> None:
        self.base_dir = base_dir
        self.args = args
        self.venv = args.venv
        self.top_dir = Path(tempfile.mkdtemp(prefix="andebox."))

    @property
    @abstractmethod
    def ansible_test(self) -> str:
        pass

    @property
    @abstractmethod
    def sub_dir(self) -> str:
        pass

    @property
    def full_dir(self) -> Path:
        return self.top_dir / self.sub_dir

    @property
    @abstractmethod
    def sanity_test_subdir(self) -> Path:
        pass

    @abstractmethod
    def post_sub_dir(self, top_dir):
        pass

    def copy_tree(self, full_dir):
        # copy files to tmp ansible coll dir
        with os.scandir() as it:
            for entry in it:
                if any(entry.name.startswith(x) for x in toplevel_exclusion):
                    continue
                if entry.is_dir():
                    shutil.copytree(
                        entry.name,
                        os.path.join(full_dir, entry.name),
                        symlinks=True,
                        ignore_dangling_symlinks=True,
                    )
                else:
                    shutil.copy(
                        entry.name,
                        os.path.join(full_dir, entry.name),
                        follow_symlinks=False,
                    )

    @contextmanager
    def temp_tree(self):
        self.full_dir.mkdir(parents=True, exist_ok=True)
        print(f"directory  = {self.full_dir}", file=sys.stderr)
        self.copy_tree(self.full_dir)

        self.post_sub_dir(self.top_dir)

        yield self.full_dir

        if self.args.keep:
            print(f"Keeping temporary directory: {self.full_dir}")
        else:
            print(f"Removing temporary directory: {self.full_dir}")
            shutil.rmtree(self.top_dir)

    def copy_exclude_lines(self, src, dest, exclusion_filenames):
        with open(src, "r") as src_file, open(dest, "w") as dest_file:
            for line in src_file.readlines():
                if not any(line.startswith(f) for f in exclusion_filenames):
                    dest_file.write(line)

    def binary_path(self, binary) -> str:
        if self.args.venv:
            return str(Path(self.args.venv) / "bin" / binary)

        return str(Path(binary))

    def exclude_from_ignore(self):
        files = [f for f in self.args.ansible_test_params if os.path.isfile(f)]
        print(f"Excluding from ignore files: {files}")
        if self.args.exclude_from_ignore:
            src_dir = Path.cwd() / self.sanity_test_subdir
            dest_dir = self.full_dir / self.sanity_test_subdir
            with os.scandir(src_dir) as ts_dir:
                for ts_entry in ts_dir:
                    if ts_entry.name.startswith("ignore") and ts_entry.name.endswith(
                        ".txt"
                    ):
                        self.copy_exclude_lines(
                            os.path.join(src_dir, ts_entry.name),
                            os.path.join(dest_dir, ts_entry.name),
                            files,
                        )


class AnsibleCoreContext(AbstractContext):
    @property
    def ansible_test(self):
        return str(self.top_dir / "bin" / "ansible-test")

    @property
    def sub_dir(self):
        return ""

    def post_sub_dir(self, top_dir):
        pass

    @property
    def sanity_test_subdir(self):
        return Path("test") / "sanity"


class CollectionContext(AbstractContext):
    def __init__(self, base_dir, args) -> None:
        super().__init__(base_dir, args)
        self.namespace, self.collection = self.determine_collection(
            self.args.collection
        )

    @property
    def ansible_test(self):
        return self.binary_path("ansible-test")

    @property
    def sub_dir(self):
        coll_dir = Path("ansible_collections") / self.namespace / self.collection
        return coll_dir

    def post_sub_dir(self, top_dir):
        print(f"collection = {self.namespace}.{self.collection}", file=sys.stderr)
        os.putenv(
            "ANSIBLE_COLLECTIONS_PATH",
            ":".join(
                [str(top_dir)]
                + os.environ.get("ANSIBLE_COLLECTIONS_PATH", "").split(":")
            ),
        )

    def install_requirements(self):
        subprocess.run(
            [
                self.binary_path("ansible-galaxy"),
                "collection",
                "install",
                "-r",
                f"{Path('tests') / 'integration' / 'requirements.yml'}",
            ],
            check=True,
        )

    @property
    def sanity_test_subdir(self):
        return str(Path("tests") / "sanity")

    def read_coll_meta(self):
        with open("galaxy.yml") as galaxy_meta:
            meta = yaml.safe_load(galaxy_meta)
        self.namespace, self.name, self.version = (
            meta["namespace"],
            meta["name"],
            meta["version"],
        )
        return meta["namespace"], meta["name"], meta["version"]

    def determine_collection(self, coll_arg):
        if coll_arg:
            coll_split = coll_arg.split(".")
            return ".".join(coll_split[:-1]), coll_split[-1]
        return self.read_coll_meta()[:2]


def _base_dir_type(dir_):
    if (dir_ / "bin" / "ansible-playbook").exists():
        return AnsibleCoreContext
    if (dir_ / "meta" / "runtime.yml").exists():
        return CollectionContext
    raise ValueError()


def _determine_base_dir_rec(
    dir_: Path,
):
    try:
        return dir_, _base_dir_type(dir_)
    except ValueError:
        if dir_ == dir_.anchor:
            raise AndeboxUnknownContext()
        return _determine_base_dir_rec(dir_.parent)


def _determine_base_dir():
    cur_dir = Path.cwd()
    try:
        return _determine_base_dir_rec(cur_dir)
    except AndeboxUnknownContext as e:
        raise AndeboxUnknownContext(
            f"Cannot determine current directory context: f{cur_dir}"
        ) from e


class Context:
    @staticmethod
    def create(args) -> AbstractContext:
        base_dir, basedir_type = _determine_base_dir()
        return (basedir_type)(base_dir, args)
